###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .graph import PropertyNode, SettableNode, CallableNode, PropSubGraphNode
from .graph import GraphError

from .. import depgraph as onyx_dg

import functools


# -------------------------------------------------------------------------
def get_node_val(caller, obj, attr, *args, **kwds):
    """
    Inputs:
        caller - caller node identity
        obj    - the instance of the ufo class or its name
        attr   - name of a stored attribute or of a method decorated by
                 @GraphNodeDescriptor()
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The value of the node.
    """    
    # --- define target node id tuple
    if isinstance(obj, str):
        target = (obj, attr, args)
    else:
        target = (obj.Name, attr, args)

    # --- get the target node from the graph (the node is created and added
    #     to the graph if it doesn't exsist)
    #     NB: this will raise the following exceptions BEFORE changing the
    #         state of the graph:
    #         - ObjNotFound    if obj is not found in database or memory
    #         - AttributeError if obj doesn't have a the required attribute
    target_node = onyx_dg.active_graph.get_or_create(target)

    # --- add target to the children set of the caller
    onyx_dg.active_graph.add_child(caller, target)

    # --- now that the graph is set up, we return the value of the target
    #     node.
    return target_node.get_value(*args, **kwds)


###############################################################################
class BaseNodeDesc(object):
    pass


###############################################################################
class PropertyNodeDesc(BaseNodeDesc):
    # -------------------------------------------------------------------------
    def __init__(self, fget):
        # --- ensure that the decorated method has the following signature:
        #         f(self, graph)
        if fget.__code__.co_argcount == 2:
            self.fget = fget
        elif fget.__code__.co_argcount < 2:
            raise GraphError(
                "Missing self and/or graph in the method definition")
        elif fget.__code__.co_argcount > 2:
            raise GraphError(
                "A 'Property' node cannot accept "
                "extra arguments besides self and graph")

    # -------------------------------------------------------------------------
    def __get__(self, instance, cls=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), attr) works as
            #     expected
            return self
        caller = (instance.Name, self.fget.__name__, ())
        return self.fget(instance, functools.partial(get_node_val, caller))

    # -------------------------------------------------------------------------
    def __set__(self, instance, value):
        raise AttributeError("Cannot set a Property node")

    # -------------------------------------------------------------------------
    def node(self, attr, ref, args):
        return PropertyNode(attr, ref)


###############################################################################
class SettableNodeDesc(BaseNodeDesc):
    # -------------------------------------------------------------------------
    def __init__(self, fget, fset, name):
        self.__name__ = name
        # ---ensure that getter and setter functions have the following
        #     signature:
        #         fget(self, graph), fset(self, graph, value)
        if fget.__code__.co_argcount == 2:
            self.fget = fget
        elif fget.__code__.co_argcount < 2:
            raise GraphError(
                "Missing self and/or graph in the getter definition")
        elif fget.__code__.co_argcount > 2:
            raise GraphError(
                "A 'Settable' node getter cannot accept "
                "extra arguments besides self and graph")
        if fset.__code__.co_argcount == 3:
            self.fset = fset
        elif fset.__code__.co_argcount < 3:
            raise GraphError(
                "Missing self and/or graph "
                "and/or value in the setter definition")
        elif fset.__code__.co_argcount > 3:
            raise GraphError(
                "A 'Settable' node setter cannot accept "
                "extra arguments besides self, graph, and value")

    # -------------------------------------------------------------------------
    def __get__(self, instance, cls=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), attr) works as
            #     expected
            return self
        caller = (instance.Name, self.__name__, ())
        return self.fget(instance, functools.partial(get_node_val, caller))

    # -------------------------------------------------------------------------
    def __set__(self, instance, value):
        caller = (instance.Name, self.__name__, ())
        self.fset(instance, functools.partial(get_node_val, caller), value)

    # -------------------------------------------------------------------------
    def node(self, attr, ref, args):
        return SettableNode(attr, ref)


###############################################################################
class CallableNodeDesc(BaseNodeDesc):
    """
    Description:
        Descriptor used for callable graph nodes, i.e. methods accepting input
        arguments.
        For callable graph nodes the input arguments are part of the node_id so
        that the same method called with different arguments will lead to
        different nodes in the graph.
        The limitation is that only positional arguments are supported.
    Signature, example:
        func(self, graph, x, y, z=1, ...)
    """
    # -------------------------------------------------------------------------
    def __init__(self, func):
        if func.__code__.co_argcount < 3:
            raise GraphError(
                "Callable node's signature requires a "
                "minimum of one argument besides self and graph")
        self.func = func

    # -------------------------------------------------------------------------
    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    # -------------------------------------------------------------------------
    def __set__(self, instance, value):
        raise AttributeError("Cannot set a Callable node")

    # -------------------------------------------------------------------------
    def __call__(self, *args):
        caller = (self.instance.Name, self.func.__name__, args)
        return self.func(
            self.instance, functools.partial(get_node_val, caller), *args)

    # -------------------------------------------------------------------------
    def node(self, attr, ref, args):
        return CallableNode(attr, ref, args)


###############################################################################
class PropSubGraphNodeDesc(BaseNodeDesc):
    """
    Description:
        Descriptor used to represent a subgraph with property-like leaf nodes.
        The input arguments are treated as children of the node so that when
        called with a different set of arguments the node in the graph will be
        invalidated.
        The limitation is that only keyword arguments are supported.
    Signature, example:
        func(self, graph, x=1, y=None, ...)
    """
    # -------------------------------------------------------------------------
    def __init__(self, func):
        argcount = func.__code__.co_argcount
        if argcount < 3:
            raise GraphError(
                "PropSubGraph node's signature requires a "
                "minimum of one argument besides self and graph")
        if not func.__defaults__ or  (argcount - 2 != len(func.__defaults__)):
            raise GraphError(
                "PropSubGraph node's signature requires all "
                "arguments besides self and graph to provide a default value")
        self.func = func

    # -------------------------------------------------------------------------
    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    # -------------------------------------------------------------------------
    def __set__(self, instance, value):
        raise AttributeError("Cannot set a PropSubGraph node")

    # -------------------------------------------------------------------------
    def __call__(self, **kwds):
        caller = (self.instance.Name, self.func.__name__, ())
        graph = functools.partial(get_node_val, caller)
        return self.func(self.instance, graph, **kwds)

    # -------------------------------------------------------------------------
    def node(self, attr, ref, args):
        return PropSubGraphNode(attr, ref)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import UseEphemeralDbs
import unittest


# -----------------------------------------------------------------------------
def main():
    with UseEphemeralDbs():
        loader = unittest.TestLoader()
        suite = loader.discover(".")
        runner = unittest.TextTestRunner(failfast=True)
        runner.run(suite)

if __name__ == "__main__":
    main()



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField

from onyx.core.depgraph.graph_api import GraphNodeDescriptor, CreateInMemory
from onyx.core.depgraph.graph_api import GetNode
from onyx.core.depgraph.graph_api import GetNodeChildren, GetNodeParents
from onyx.core.depgraph.graph_api import GetVal, InvalidateNode
from onyx.core.utils.unittest import OnyxTestCase

import unittest


###############################################################################
class Fibonacci(UfoBase):
    fib0 = IntField(default=0)
    fib1 = IntField(default=1)

    # -------------------------------------------------------------------------
    @GraphNodeDescriptor("Callable")
    def fib(self, graph, k):
        if k == 0:
            return graph(self, "fib0")
        elif k == 1:
            return graph(self, "fib1")
        else:
            return graph(self, "fib", k-1) + graph(self, "fib", k-2)


###############################################################################
class TestFibonacci(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(Fibonacci(Name="fib"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_initial_values(self):
        self.assertEqual(GetVal(self.name, "fib0"), 0)
        self.assertEqual(GetVal(self.name, "fib1"), 1)

    # -------------------------------------------------------------------------
    def test_seq(self):
        self.assertEqual(GetVal(self.name, "fib", 0), 0)
        self.assertEqual(GetVal(self.name, "fib", 1), 1)
        self.assertEqual(GetVal(self.name, "fib", 2), 1)
        self.assertEqual(GetVal(self.name, "fib", 3), 2)
        self.assertEqual(GetVal(self.name, "fib", 4), 3)
        self.assertEqual(GetVal(self.name, "fib", 5), 5)
        self.assertEqual(GetVal(self.name, "fib", 6), 8)
        self.assertEqual(GetVal(self.name, "fib", 7), 13)
        self.assertEqual(GetVal(self.name, "fib", 8), 21)
        self.assertEqual(GetVal(self.name, "fib", 9), 34)

    # -------------------------------------------------------------------------
    def test_graph(self):
        """
        Calculate a large fibonacci number and then perform tests on the
        generated graph.
        """
        self.assertEqual(GetVal(self.name, "fib", 10), 55)

        parents = {('fib', 'fib', (9,)), ('fib', 'fib', (10,))}
        children = {('fib', 'fib', (6,)), ('fib', 'fib', (7,))}

        node_id = (self.name, "fib", (8,))
        node = GetNode(*node_id)

        self.assertTrue(node.valid)
        self.assertEqual(node.value, 21)
        self.assertEqual(GetNodeParents(*node_id), parents)
        self.assertEqual(GetNodeChildren(*node_id), children)

        # --- invalidate node 8 and check again the state of the graph
        InvalidateNode(self.name, "fib", (8,))

        # --- invalidation sets the node to invalid and afftects the
        #     topology of the graph by resetting all children (and, therefore,
        #     all the parents)
        self.assertFalse(node.valid)
        self.assertEqual(node.value, 21)
        self.assertEqual(GetNodeParents(*node_id), set())
        self.assertEqual(GetNodeChildren(*node_id), set())


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.date import Date
from ..datatypes.gcurve import GCurve
from ..datatypes.curve import Curve
from ..database.ufo_base import UfoBase, BaseField, custom_decoder
from ..database.ufo_fields import FloatField
from ..database.objdb import ObjNotFound
from ..database.objdb_api import ObjDbQuery
from .graph_api import GetVal, GraphNodeDescriptor

from .. import database as onyx_db

import json

__all__ = [
    "Archivable",
    "MktIndirectionFactory",
    "EnforceArchivableEntitlements",
]

GET_HISTORY_QUERY = """
SELECT Date, Value FROM Archive
WHERE Name=%s AND Attribute=%s AND Date BETWEEN %s AND %s ORDER BY Date;"""

GET_MOST_RECENT = """
SELECT MAX(Date) FROM Archive
WHERE Name=%s AND Attribute=%s AND Date <= %s"""


# -----------------------------------------------------------------------------
def empty():
    pass

def empty_with_docstring():
    """Empty function with docstring."""
    pass

BYTECODE_OF_EMPTY = empty.__code__.co_code
BYTECODE_OF_EMPTY_WITH_DOCSTRING = empty_with_docstring.__code__.co_code

def has_empty_body(func):
    """
    Helper function used to tell if a function has an empty body.
    """
    if func.__doc__:
        return func.__code__.co_code == BYTECODE_OF_EMPTY_WITH_DOCSTRING
    else:
        return func.__code__.co_code == BYTECODE_OF_EMPTY


###############################################################################
class Archivable(UfoBase):
    """
    Base class for all archivable UFO objects.
    """

    # -------------------------------------------------------------------------
    def last_before(self, attr, date=None):
        """
        Description:
            Return the date of the most recent archived record before a given
            date.
        Inputs:
            attr - the required attribute
            date - look for dated records with date earlier than this value.
        Returns:
            A Date.
        """
        date = date or Date.high_date()
        res = ObjDbQuery(
            GET_MOST_RECENT, (self.Name, attr, date), attr="fetchone")
        if res.max is None:
            raise ObjNotFound(
                "{0!s} not found for any date earlier"
                "than {1:s}".format((self.Name, attr), date))
        else:
            return res.max

    # -------------------------------------------------------------------------
    def get_dated(self, attr, date, strict=False, refresh=False):
        """
        Description:
            Get the value of an archived attribute as of a given date.
        Inputs:
            attr    - the attribute to archive
            date    - the archive date
            strict  - if true, raise an ObjNotFound exception if a record
                      cannot be found for the required date
            refresh - if true, force reloading archived record from backend
        Returns:
            A (Date, value) tuple
        """
        # --- get the instance of the field descriptor that is storing
        #     the attribute descriptor itself
        field = getattr(self.__class__, attr).field

        # --- load from database and de-serialize value
        date, value = onyx_db.obj_clt.get_dated(
            self.Name, attr, date, strict=strict, refresh=refresh)

        return date, field.from_json(value)

    # -------------------------------------------------------------------------
    def set_dated(self, attr, date, value, overwrite=False):
        """
        Description:
            Set the value of an archived attribute.
        Inputs:
            attr      - the attribute to archive
            date      - the archive date
            value     - the value to archive
            overwrite - if true, existing records can be overwritten
        """
        # --- get the instance of the field descriptor that is storing
        #     the attribute descriptor itself
        field = getattr(self.__class__, attr).field
        # --- make sure value is compatible with the field_type
        field.validate(value)
        # --- serialize value
        value = field.to_json(value)

        onyx_db.obj_clt.set_dated(
            self.Name, attr, date, value, overwrite=overwrite)

    # -------------------------------------------------------------------------
    def delete(self):
        """
        Invoked when the archived object is deleted from database.
        """
        ObjDbQuery("DELETE FROM Archive WHERE Name=%s", parms=(self.Name,))

    # -------------------------------------------------------------------------
    def get_history(self, attr, start=None, end=None):
        start = start or Date.low_date()
        end = end or Date.high_date()

        # --- get the instance of the field descriptor that is stored
        #     the VT descriptor itself
        field = getattr(self.__class__, attr).field

        parms = (self.Name, attr, start, end)
        rows = ObjDbQuery(GET_HISTORY_QUERY, parms, attr="fetchall")

        # --- usually json.loads is called by get_dated before returning the
        #     value but here we are processing results from a raw query
        def convert(val):
            return field.from_json(json.loads(val, cls=custom_decoder))

        knots = [(Date.parse(r[0]), convert(r[1])) for r in rows]

        if isinstance(field, FloatField):
            return Curve([d for d, v in knots], [v for d, v in knots])
        else:
            return GCurve([d for d, v in knots], [v for d, v in knots])


###############################################################################
class MktIndirectionFactory(object):
    """
    Description:
        Factory decorator that implements a descriptor protocol used for market
        data indirection.
    Inputs:
        field_type - a valid field-type constructor (must be a subclass of
                     BaseField).
        obj        - name of the object exposing value types with the current
                     market data date and a flag dictating if archived values
                     need to be looked up in strict mode or not (default is
                     "Database").
        mdd_attr   - name of the attribute for market data date (default
                     is "MktDataDate").
        fs_attr    - name of the attribute for force strict flag (default
                     is "ForceStrict").

    THIS DECORATOR CAN ONLY BE APPLIED TO METHODS OF A SUBCLASS OF Archivable.

    Typical use is as follows:
        @MktIndirectionFactory(FloatField)
        def MarketizedAttr(self, graph):
            pass

    or:
        @MktIndirectionFactory(FloatField)
        def MarketizedAttr(self, graph):
            default implementation...

    MarketizedAttr is created as a Property node descriptor that fetches
    archived data by accessing the corresponding dated record and, optionally,
    defaults (in case of missing dated record) to executing the function
    itself.
    Type consistency with the field type is enforced.
    """
    # -------------------------------------------------------------------------
    def __init__(self, field_type, obj="Database",
                 mdd_attr="MktDataDate", fs_attr="ForceStrict"):
        if not issubclass(field_type, BaseField):
            raise ValueError(
                "field_type must be an instance of a class derived from "
                "BaseField. Got {0!s} instead".format(field_type.__class__))

        self.field = field_type()
        self.obj = obj
        self.mdd_attr = mdd_attr
        self.fs_attr = fs_attr

    # -------------------------------------------------------------------------
    def __call__(self, func):
        func_name = func.__name__
        func_empty = has_empty_body(func)

        # --- descriptor protocol: getter
        def getter(instance, graph):
            mdd = graph(self.obj, self.mdd_attr)  # market data date
            fs = graph(self.obj, self.fs_attr)  # force strict
            try:
                _, value = instance.get_dated(func_name, mdd, fs)
            except ObjNotFound:
                if func_empty:
                    raise
                else:
                    value = func(instance, graph)
            return value

        getter.__name__ = func_name

        # --- create a property descriptor and add the appropriate
        #     field attribute to it
        descriptor = GraphNodeDescriptor()(getter)
        descriptor.field = self.field

        return descriptor


###############################################################################
class EnforceArchivableEntitlements(object):
    """
    Description:
        Decorator class used to enforce compliance with Archivable
        entitlements.
    Inputs:
        obj  - target object as defined in database (or memory)
        attr - target attribute (a stored attribute or a method of a ufo
               class decorated by @GraphNodeDescriptor())

    Typical use is as follows:

        @EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
        class my_ufo_class(Archivable):
        ...
    """
    # -------------------------------------------------------------------------
    def __init__(self, obj, attr):
        self.obj = obj
        self.attr = attr

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        if not issubclass(cls, Archivable):
            raise ValueError(
                "enforce_archivable_entitlements can "
                "only be applied to a subclass of Archivable")

        # --- override set_dated to enforce compliance with the required value
        #     type
        def set_dated(instance, attr, date, value):
            overwrite = GetVal(self.obj, self.attr)
            instance.__set_dated_raw(attr, date, value, overwrite)

        # --- rename original set_dated and replace it with a new version
        cls.__set_dated_raw = cls.set_dated
        cls.set_dated = set_dated

        return cls



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import collections
import copy
import weakref


###############################################################################
class GraphError(Exception):
    pass


###############################################################################
class DependencyGraph(object):
    """
    Class representing the onyx Dependency Graph (a Directed Acyclic Graph).
    Current implementation uses a dictionary that maps tuple of the type
    (ObjectName, AttributeName, args=()) to graph nodes (instances of
    GraphNode class).
    """
    # -------------------------------------------------------------------------
    def __init__(self, db_clt):
        # --- database client instance
        self.db_clt = db_clt
        # --- here we store the nodes
        self.nodes = dict()
        # --- for each node we keep the list of all its children
        self.children = collections.defaultdict(lambda: set())
        # --- for each node we keep the list of all its parents
        self.parents = collections.defaultdict(lambda: set())

    # -------------------------------------------------------------------------
    def add_child(self, node_id, child):
        self.children[node_id].add(child)
        self.parents[child].add(node_id)

    # -------------------------------------------------------------------------
    def clear_children(self, node_id):
        for child in self.children[node_id]:
            # --- this should never raise KeyError
            self.parents[child].remove(node_id)
        self.children[node_id].clear()

    # -------------------------------------------------------------------------
    def get_or_create(self, node_id):
        try:
            return self.nodes[node_id]
        except KeyError:
            self.nodes[node_id] = node = self.create_node(node_id)
            return node

    # -------------------------------------------------------------------------
    def create_node(self, node_id):
        if len(node_id) == 2:
            obj_name, attr = node_id
            args = ()
        else:
            obj_name, attr, args = node_id
    
        obj = self.db_clt.get(obj_name)
    
        if attr == "StoredAttrs":
            # --- StoredAttrs is special insofar that we don't want it to be
            #     settable/changeable
            return GraphNode(attr, weakref.ref(obj))
        elif attr in obj.StoredAttrs:
            # --- all stored attributes are settable by definition
            return SettableNode(attr, weakref.ref(obj))
        else:
            # --- use the node constructor of the node descriptor itself
            node_descriptor = getattr(obj.__class__, attr)
            return node_descriptor.node(attr, weakref.ref(obj), args)

    # -------------------------------------------------------------------------
    def invalidate_node(self, node_id):
        try:
            for parent in self.parents[node_id].copy():
                self.invalidate_node(parent)

            self.nodes[node_id].valid = False
            self.clear_children(node_id)
        except KeyError:
            # --- early return if node doesn't exist
            return

    # -------------------------------------------------------------------------
    def clear(self):
        self.nodes.clear()
        self.children.clear()
        self.parents.clear()

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__new__(self.__class__)
        clone.db_clt = self.db_clt
        clone.nodes = copy.deepcopy(self.nodes)
        clone.children = self.children.copy()
        clone.parents = self.parents.copy()
        memo[id(self)] = clone
        return clone


###############################################################################
class GraphNode(object):
    """
    Base class representing a node in the Dependency Graph.
    """
    __slots__ = ("attr", "obj_ref", "valid", "value")

    # -------------------------------------------------------------------------
    def __init__(self, attr, obj_ref):
        self.attr = attr
        self.obj_ref = obj_ref
        self.valid = False
        self.value = None

    # -------------------------------------------------------------------------
    def get_id(self):
        return self.obj_ref().Name, self.attr, ()

    # -------------------------------------------------------------------------
    def get_value(self):
        if not self.valid:
            # --- get the attribute's value from the instance itself and return
            #     it to the caller
            self.value = getattr(self.obj_ref(), self.attr)
            self.valid = True

        return copy.deepcopy(self.value)

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__new__(self.__class__)
        clone.attr = self.attr
        clone.obj_ref = self.obj_ref
        clone.valid = self.valid
        # --- value can be an instance of a mutable class, make a proper copy.
        clone.value = copy.deepcopy(self.value)
        memo[id(self)] = clone
        return clone


###############################################################################
class PropertyNode(GraphNode):
    """
    Derived class used for Property ValueTypes.
    """
    __slots__ = ()


###############################################################################
class SettableNode(GraphNode):
    """
    Derived class used for nodes whose value can be explicitly set (such as
    StoredAttrs and descriptors that implement a setter).
    """
    __slots__ = ()


###############################################################################
class CallableNode(GraphNode):
    __slots__ = ("args",)

    # -------------------------------------------------------------------------
    def __init__(self, attr, obj_ref, args):
        super().__init__(attr, obj_ref)
        self.args = args

    # -------------------------------------------------------------------------
    def get_id(self):
        return self.obj_ref().Name, self.attr, self.args

    # -------------------------------------------------------------------------
    def get_value(self, *args):
        assert args == self.args
        if not self.valid:
            # --- call the object's method with the arguments for this node
            self.value = getattr(self.obj_ref(), self.attr)(*args)
            self.valid = True

        return copy.deepcopy(self.value)

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = super().__deepcopy__(memo)
        # --- shallow copies here should be fine
        clone.args = self.args[:]
        memo[id(self)] = clone
        return clone


###############################################################################
class PropSubGraphNode(GraphNode):
    __slots__ = ("kwds",)

    # -------------------------------------------------------------------------
    def __init__(self, attr, obj_ref):
        super().__init__(attr, obj_ref)
        self.kwds = {}

    # -------------------------------------------------------------------------
    def get_value(self, **kwds):
        # --- if kwds don't match the stored values, the node is assumed to be
        #     no longer valid
        if kwds != self.kwds:
            self.valid = False
            self.kwds = kwds

        if not self.valid:
            # --- call the object's method with provided arguments
            self.value = getattr(self.obj_ref(), self.attr)(**kwds)
            self.valid = True

        return copy.deepcopy(self.value)

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = super().__deepcopy__(memo)
        clone.kwds = copy.deepcopy(self.kwds)
        memo[id(self)] = clone
        return clone



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import ctypes
import subprocess
import sys
import os
import logging
import argh

SEM_FAILCRITICALERRORS = 1
SEM_NOGPFAULTERRORBOX  = 2
SEM_NOOPENFILEERRORBOX = 32768 # value from win32con.py
NEW_SETTINGS = (
    SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX)

# -----------------------------------------------------------------------------
def run_forever(
    executable=None, script=None, logging_config_file=None, timeout=None):

    # --- if not specified, it defaults to the python executable
    executable = executable or sys.executable

    if script is None:
        args = [executable]
    else:
        args = [executable, script]

    if logging_config_file is not None:
        if os.path.exists(logging_config_file):
            args.append("-l")
            args.append(logging_config_file)
        else:
            raise FileNotFoundError("Logging "
                "config file {0:s} not found".format(logging_config_file))

    # --- configure logger for the run_forever script
    config = {
        "level": logging.INFO,
        "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    }
    logging.basicConfig(**config)
    logger = logging.getLogger(__name__)

    if sys.platform == "win32":
        # --- don't display the Windows GPF dialog if the invoked program dies
        #     as per:
        #      - https://mail.python.org/pipermail/python-list/2009-January/520255.html
        #     see also:
        #      - https://blogs.msdn.microsoft.com/oldnewthing/20040727-00/?p=38323
        #      - https://www.activestate.com/blog/2007/11/supressing-windows-error-report-messagebox-subprocess-and-ctypes
        old_settings = ctypes.windll.kernel32.SetErrorMode(NEW_SETTINGS)
        ctypes.windll.kernel32.SetErrorMode(old_settings | NEW_SETTINGS)

    while True:
        try:
            subprocess.check_call(
                args, shell=False, timeout=timeout, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as err:
            logger.error("subprocess died with error: {0!s}".format(err))
        except subprocess.TimeoutExpired:
            logger.info("subprocess reached timeout, restarting...")
        else:
            logger.info("subprocess finished without errors, restarting...")


# -----------------------------------------------------------------------------
def main():
    argh.dispatch_command(run_forever)
        


###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .graph_api import GraphNodeDescriptor

import types

__all__ = [
    "RetainedFactory",
    "InheritAsProperty",
    "DiscardInheritedAttribute",
]


###############################################################################
class RetainedFactory(object):
    """
    Description:
        GraphNodeDescriptor factory that implements a descriptor protocol used
        to represent pseudo-attribute nodes, i.e. nodes that can be set but are
        not persisted in database.

    Typical use cases are for the implementation of retained properties (such
    as Spot) where the syntax is as follows:

        @RetainedFactory()
        def Spot(self, graph):
            ...
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        self.values = {}

    # -------------------------------------------------------------------------
    def __call__(self, func):
        def getter(instance, graph):
            try:
                return self.values[instance.Name]
            except KeyError:
                return func(instance, graph)

        def setter(instance, graph, value):
            self.values[instance.Name] = value

        # --- return a settable descriptor ValueType
        return GraphNodeDescriptor("Settable", getter, setter)(func)


###############################################################################
class InheritAsProperty(object):
    """
    Description:
        Class decorator used to replace one or more stored attributes of the
        super class with properties pointing to the same attribute of another
        object.
    Inputs:
        attrs - a list of stored attributes of the super-class that are to be
                replaced by Property value types.
        ptr   - a pointer to the instance of the ufo class from which we are
                inheriting stored attributes as properties.
                Each stored attribute in attrs will be replaced with by
                property node.
    """
    template = ("def template(self, graph):\n"
                "    return graph(graph(self, '{0:s}'), '{1:s}')")

    # -------------------------------------------------------------------------
    def __init__(self, attrs, ptr):
        self.attrs = attrs
        self.ptr = ptr

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        for attr in self.attrs:
            # --- discard from set of StoredAttrs attributes (discard does
            #     nothing if cls doesn't have such attribute)
            cls._json_fields.discard(attr)
            cls.StoredAttrs.discard(attr)

            # --- create Property-NodeDescriptor
            mod = types.ModuleType("__templates")
            exec(self.template.format(self.ptr, attr), mod.__dict__)
            func = types.FunctionType(mod.template.__code__, {}, name=attr)
            setattr(cls, attr, GraphNodeDescriptor()(func))

        return cls


###############################################################################
class DiscardInheritedAttribute(object):
    """
    Description:
        Class decorator used to dis-inherit one or more stored attributes of
        the super class, ususally so that they can be re-implemented as
        properties.
    Inputs:
        attrs - a list of stored attributes of the super-class that are
                to be discarded.
    """
    # -------------------------------------------------------------------------
    def __init__(self, attrs):
        self.attrs = attrs

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        for attr in self.attrs:
            # --- discard from set of StoredAttrs attributes (discard does
            #     nothing if cls doesn't have such attribute)
            cls._json_fields.discard(attr)
            cls.StoredAttrs.discard(attr)

        return cls



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..database.ufo_base import get_base_classes
from ..database.objdb import ObjNotFound
from ..database.objdb_api import DelObj
from .graph import DependencyGraph, GraphError, SettableNode, CallableNode
from .graph_node_descriptors import BaseNodeDesc, PropertyNodeDesc
from .graph_node_descriptors import SettableNodeDesc, CallableNodeDesc
from .graph_node_descriptors import PropSubGraphNodeDesc

from .. import database as onyx_db
from .. import depgraph as onyx_dg

__all__ = [
    "GraphError",
    "UseGraph",
    "GraphNodeDescriptor",
    "CreateInMemory",
    "RemoveFromGraph",
    "PurgeObj",
    "GetVal",
    "SetVal",
    "GetNode",
    "GetNodeChildren",
    "GetNodeParents",
    "InvalidateNode",
    "IsInstance",
    "ChildrenSet",
]


###############################################################################
class UseGraph(object):
    """
    Context manager used to created and activate the dependency graph.
    """
    def __init__(self, graph=None):
        self.graph = graph or DependencyGraph(db_clt=onyx_db.obj_clt)

    def __enter__(self):
        # --- make sure the database client is available
        if self.graph.db_clt is None:
            raise GraphError("global database "
                "client is not available (did you call UseDatabase?)")

        # --- we don't support nested graphs as it would make the API
        #     implementation more complicated.
        #     to isolate the active graph from changes GraphScope should be
        #     used instead
        if onyx_dg.active_graph is not None:
            raise GraphError(
                "Nested dependency graphs are "
                "not supported, use a GraphScope instead")

        # --- set the active graph
        onyx_dg.active_graph = self.graph

    def __exit__(self, *args, **kwds):
        onyx_dg.active_graph = None
        # --- returns False so that all execptions raised will be propagated
        return False


###############################################################################
class GraphNodeDescriptor(object):
    """
    This decorator trasforms methods of classes derived from UfoBase into a
    graph node descriptor which in turn is used to create the nodes of the
    DependencyGraph.
    """
    # -------------------------------------------------------------------------
    def __init__(self, node_type="Property", fget=None, fset=None):
        self.node_type = node_type
        if node_type == "Settable":
            if fget is None or fset is None:
                raise GraphError("Settable nodes need both getter and setter")
            else:
                self.fget = fget
                self.fset = fset

    # -------------------------------------------------------------------------
    def __call__(self, func):
        if self.node_type == "Property":
            return PropertyNodeDesc(func)
        elif self.node_type == "Settable":
            return SettableNodeDesc(self.fget, self.fset, func.__name__)
        elif self.node_type == "Callable":
            return CallableNodeDesc(func)
        elif self.node_type == "PropSubGraph":
            return PropSubGraphNodeDesc(func)
        else:
            raise GraphError(
                "Unrecognized node type {0:s}".format(self.node_type))


# -----------------------------------------------------------------------------
def CreateInMemory(instance):
    """
    Description:
        Helper function that adds an instance of an object derived from
        UfoBase in memory so that all its attributes and decorated methods
        are visible to the DependencyGraph and their values can be obtained
        using the standard graph API.
    Inputs:
        instance - the instance of an object derived from UfoBase
    Returns:
        The instance itself.
    """
    if instance.Name is None:
        raise GraphError("instance.Name was not set")

    try:
        # --- return a reference to the cached instance with same name
        return onyx_db.obj_instances[instance.Name]
    except KeyError:
        # --- add object instance to global cache
        onyx_db.obj_instances[instance.Name] = instance
        # --- return a reference to the instance
        return instance


# -----------------------------------------------------------------------------
def RemoveFromGraph(obj):
    """
    Description:
        Remove all nodes referencing a given object from the graph.
    Inputs:
        obj - instance (or name) of an object derived from UfoBase
    Returns:
        None.
    """
    obj_name = obj if isinstance(obj, str) else obj.Name

    # --- list of all nodes ids referencing the given object
    nodes = [node_id
        for node_id, node in onyx_dg.active_graph.nodes.items()
        if node_id[0] == obj_name]

    # --- invalidate all nodes referencing the given object
    for node_id in nodes:
        onyx_dg.active_graph.invalidate_node(node_id)

    # --- remove from the graph all nodes referencing the given object
    for node_id in nodes:
        del onyx_dg.active_graph.nodes[node_id]


# -----------------------------------------------------------------------------
def PurgeObj(obj):
    """
    Description:
        Remove all nodes referencing a given object from the graph and then
        delete the object itself from the database.
    Inputs:
        obj - an instance of an object derived from UfoBase (or its name)
    Returns:
        None.
    """
    RemoveFromGraph(obj)
    try:
        DelObj(obj)
    except ObjNotFound:
        pass


# -----------------------------------------------------------------------------
def GetVal(obj, attr, *args, **kwds):
    """
    Description:
        Get/calculate the value of a graph node. If the node doesn't currently
        exist, it will be created.
        NB: this function should never be used when writing ufo classes as
            it doesn't 'wire' caller and target together, leaving their
            relationship undefined. The injected graph handler should be used
            instead.
    Inputs:
        obj    - instance (or name) of an object derived from UfoBase
        attr   - target's attribute/method
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The graph node's value.
    """
    name = obj if isinstance(obj, str) else obj.Name
    node = onyx_dg.active_graph.get_or_create((name, attr, args))
    return node.get_value(*args, **kwds)


# -----------------------------------------------------------------------------
def SetVal(obj, attr, value):
    """
    Description:
        Set the value of a stored attribute on-graph (i.e. the new value is
        visible on the graph). When the instance is persisted by calling
        UpdateObj(obj.Name), the new attribute's value is persisted as well.
        NB: only settable nodes (generally stored attributes) can be set.
    Inputs:
        obj   - instance (or name) of an object derived from UfoBase
        attr  - target's attribute/method
        value - the value to set the VT to
    Returns:
        None.
    """
    name = obj if isinstance(obj, str) else obj.Name

    # --- get the target node from the graph
    node_id = (name, attr, ())
    node = onyx_dg.active_graph.get_or_create(node_id)

    if not isinstance(node, SettableNode):
        raise GraphError(
            "({0!s}, {1!s}, {2!s}) is not a settable node".format(*node_id))

    # --- remove all its children and invalidate recursively the node and its
    #     ancestors
    onyx_dg.active_graph.clear_children(node_id)
    onyx_dg.active_graph.invalidate_node(node_id)

    # --- set the node value and set its state to valid
    node.value = value
    node.valid = True

    # --- set attribute in the cached instance of the object
    setattr(node.obj_ref(), attr, value)


# -----------------------------------------------------------------------------
def GetNode(obj, attr, args=()):
    """
    Description:
        Returns a node from the graph if present, otherwise it creates the new
        node and adds it to the graph.
        NB: application code should never need to fetch a node directly, but
            rather use GetVal.
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        attr - name of the target attribute/method
        args - tuple of arguments defining a callable node (optional)
    Returns:
        A reference to the node in the graph.
    """
    name = obj if isinstance(obj, str) else obj.Name
    return onyx_dg.active_graph.get_or_create((name, attr, args))


# -----------------------------------------------------------------------------
def GetNodeChildren(obj, attr, args=()):
    """
    Description:
        Returns the set of children of a given node.
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        attr - name of the target attribute/method
        args - tuple of arguments defining a callable node (optional)
    Returns:
        A set.
    """
    name = obj if isinstance(obj, str) else obj.Name
    return onyx_dg.active_graph.children[(name, attr, args)]


# -----------------------------------------------------------------------------
def GetNodeParents(obj, attr, args=()):
    """
    Description:
        Returns the set of parents of a given node.
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        attr - name of the target attribute/method
        args - tuple of arguments defining a callable node (optional)
    Returns:
        A set.
    """
    name = obj if isinstance(obj, str) else obj.Name
    return onyx_dg.active_graph.parents[(name, attr, args)]


# -----------------------------------------------------------------------------
def InvalidateNode(obj, attr, args=()):
    """
    Description:
        Invalidate a node in the graph (if it exists).
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        attr - target's attribute/method
        args - tuple of arguments defining a callable node (optional)
    Returns:
        None.
    """
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, attr, args)

    try:
        node = onyx_dg.active_graph.nodes[node_id]
    except KeyError:
        pass
    else:
        if isinstance(node, CallableNode) and args == ():
            raise GraphError(
                "To invalidate a callable node it is "
                "mandatory to specify the tuple of arguments")

        onyx_dg.active_graph.invalidate_node(node_id)


## -----------------------------------------------------------------------------
#def NodesByObjName(obj):
#    """
#    Description:
#        Returns a list of all the nodes that reference the given object.
#    Inputs:
#        obj - instance (or name) of an object derived from UfoBase
#    Yields:
#        A list of Graph nodes.
#    """
#    obj_name = obj if isinstance(obj, str) else obj.Name
#    
#    return [node
#        for node_id, node in onyx_dg.active_graph.items()
#        if node_id[0] == obj_name]


# # -----------------------------------------------------------------------------
# def ValueTypesByInstance(obj):
#     """
#     Description:
#         Returns a list of all value types for object obj
#     Inputs:
#         obj - instance (or name) of an object derived from UfoBase
#     Returns:
#         A list of VT names.
#     """
#     if isinstance(obj, str):
#         obj = onyx_db.obj_clt.get(obj)
#
#     cls = obj.__class__
#     calc = [name for name, VT in
#             cls.__dict__.items() if isinstance(VT, BaseVT)]
#
#     return list(obj.StoredAttrs) + calc


# -----------------------------------------------------------------------------
def IsInstance(name, obj_type):
    """
    Description:
        This is the equivalent of isinstance, but takes the instance name and
        the class name as inputs.
    Inputs:
        name     - the instance name
        obj_type - the class name
    """
    obj = onyx_db.obj_clt.get(name)
    bases = set(get_base_classes(obj.__class__))
    return obj_type in bases


# -----------------------------------------------------------------------------
def children_iterator(node_id, child_attr, obj_type):
    for name, attr, args in onyx_dg.active_graph.children[node_id]:
        if attr == child_attr:
            if obj_type is None:
                yield name
            else:
                obj = onyx_db.obj_instances[name]
                bases = set(get_base_classes(obj.__class__))
                if obj_type in bases:
                    yield name
        else:
            yield from children_iterator(
                (name, attr, args), child_attr, obj_type)


# -----------------------------------------------------------------------------
def ChildrenSet(node_id, child_attr, obj_type=None, graph=None):
    """
    Description:
        Given a node_id, return the names of all those objects such that the
        node (object, attr) is a descendant of node_id.
        Optionally, children can be restricted to be instances of a given
        class.
    Inputs:
        node_id    - parent node_id, in the form (name, attr, args), where args
                     can be skipped in place of an empty tuple
        child_attr - children attribute's name
        obj_type   - if set, children must be instances of this class (or of a
                     subclass)
        graph      - if set, function used to get/calculate the node's value.
    Yields:
        A set with object names.
    """
    obj, attr, *args = node_id
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, attr, tuple(args))

    if graph is None:
        GetVal(obj, attr, *args)
    else:
        graph(obj, attr, *args)

    return set(children_iterator(node_id, child_attr, obj_type))


## -----------------------------------------------------------------------------
#def leaves_iterator(node_id, get_val):
#    stored = get_val(node_id[0], "StoredAttrs")
#
#    for name, vt, args in onyx_dg.active_graph[node_id].children:
#        if vt in stored:
#            yield name, vt
#        else:
#            for leaf in leaves_iterator((name, vt, args), get_val):
#                yield leaf


## -----------------------------------------------------------------------------
#def LeafNodes(node_id, graph=None):
#    """
#    Description:
#        Return all leaf-level nodes (StoredAttrs) of a given node.
#    Inputs:
#        node_id - parent node_id, in the form (Name, VT, args), where args can
#                  be skipped in place of an empty tuple
#        graph   - if set, function used to generate graph topology
#    Yields:
#        A set of nodes.
#    """
#    obj, VT, *args = node_id
#    name = obj if isinstance(obj, str) else obj.Name
#    node_id = (name, VT, tuple(args))
#
#    get_val = graph or GetVal
#    get_val(obj, VT, *args)
#
#    return set(leaves_iterator(node_id, get_val))


## -----------------------------------------------------------------------------
#def settable_iterator(node_id):
#    for kid_id in onyx_dg.active_graph[node_id].children:
#        node = onyx_dg.active_graph[kid_id]
#        for leaf in settable_iterator(kid_id):
#            if leaf is not None:
#                yield leaf
#
#        if isinstance(node, SettableNode):
#            yield kid_id


## -----------------------------------------------------------------------------
#def SettableNodes(node_id, graph=None):
#    """
#    Description:
#        Return all settable children of a given node (stopping at the first
#        level).
#    Inputs:
#        node_id - parent node_id, in the form (Name, VT, args), where args can
#                  be skipped in place of an empty tuple
#        graph   - if set, function used to generate graph topology
#    Yields:
#        A set of nodes.
#    """
#    obj, VT, *args = node_id
#    name = obj if isinstance(obj, str) else obj.Name
#    node_id = (name, VT, tuple(args))
#
#    if graph is None:
#        GetVal(obj, VT, *args)
#    else:
#        graph(obj, VT, *args)
#
#    return set(settable_iterator(node_id))



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .graph import DependencyGraph, GraphError, PropertyNode, SettableNode

from .. import database as onyx_db
from .. import depgraph as onyx_dg

import collections
import copy

__all__ = ["EvalBlock", "GraphScope"]


###############################################################################
class EvalBlock(object):
    """
    Description:
        Context manager used to manage lifetime of one or more one-off changes.
        NB: when exiting an EvalBlock, all changed nodes as well as their
        ancestors are invalidated and the topology of the graph is lost.
    Usage:
        Typical use is as follows:

        with EvalBlock() as eb:
            eb.change_value("abc", "xyz", 123)
            ...
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        self.__changed_nodes = {}

    # -------------------------------------------------------------------------
    def __enter__(self):
        # --- return a reference to itself (to be used by change_value)
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        for node_id, original_node in self.__changed_nodes.items():
            # --- invalidate the node and all its ancestors
            onyx_dg.active_graph.invalidate_node(node_id)

            # --- move pre-change node back into the graph
            onyx_dg.active_graph.nodes[node_id] = original_node

        self.__changed_nodes.clear()

        # --- returns False so that all execptions raised will be propagated
        return False

    # -------------------------------------------------------------------------
    def change_value(self, obj, attr, value):
        """
        Description:
            Change the in-memory value of a graph node within an EvalBlock.
        Inputs:
            obj   - instance (or name) of an object derived from UfoBase
            attr  - name of the target attribute/method
            value - the new value for the attribute
        Returns:
            None.
        """
        name = obj if isinstance(obj, str) else obj.Name
        node_id = (name, attr, ())
        is_new_node = False

        try:
            node = onyx_dg.active_graph.nodes[node_id]
        except KeyError:
            node = onyx_dg.active_graph.create_node(node_id)
            onyx_dg.active_graph.nodes[node_id] = node
            is_new_node = True

        if not isinstance(node, (PropertyNode, SettableNode)):
            node_type = node.__class__.__name__
            raise GraphError(
                "Unsupported node type: {0:s}".format(node_type))

        # --- copy pre-change node and store it in __changed_nodes. This
        #     is only required the first time a node is changed.
        if node_id not in self.__changed_nodes:
            original = copy.deepcopy(node)
            if len(onyx_dg.active_graph.children[node_id]) :
                # --- this is a calculated node, invalidate it
                original.valid = False
            else:
                # --- this is a leaf or invalid node, do nothing
                pass
            self.__changed_nodes[node_id] = original

        # --- invalidation is only needed for pre-existing nodes
        if not is_new_node:
            onyx_dg.active_graph.invalidate_node(node_id)

        # --- set the node value and set its state to valid
        node.value = value
        node.valid = True


###############################################################################
class dict_with_fallback(collections.UserDict):
    # -------------------------------------------------------------------------
    def __init__(self, fallback, *args, **kwds):
        super().__init__(*args, **kwds)
        self.fallback = fallback

    # -------------------------------------------------------------------------
    def __getitem__(self, item):
        # --- return the value if present, otherwise get it from the fallback
        #     dictionary, make a deepcopy and return it
        try:
            return self.data[item]
        except KeyError:
            self.data[item] = value = copy.deepcopy(self.fallback[item])
            return value

    # -------------------------------------------------------------------------
    def __iter__(self):
        return iter({**self.data, **self.fallback})


###############################################################################
class GraphScope(DependencyGraph):
    """
    Description:
        Context manager used to manage the lifetime of in-memory changes to
        the graph. All changes applied within a GraphScope remain invisible
        outside the context manager (including graph topology).
        A GraphScope can be used to create persistent scenarions that are
        then re-used multiple times.
    Usage:
        Typical use is as follows:

        scope = GraphScope()
        scope.change_value("abc", xyz", 123)
        with scope:
            scope.change_value("abc", "xxx", 666)
            ...

        with scope:
            ...
            ...
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        super().__init__(db_clt=onyx_dg.active_graph.db_clt)
        # --- save a reference to the underlying graph
        self.udl_graph = onyx_dg.active_graph
        # --- we override nodes, children and parents with custom dictionaries
        #     which fallback to the underlying graph
        self.nodes = dict_with_fallback(self.udl_graph.nodes)
        self.children = dict_with_fallback(self.udl_graph.children)
        self.parents = dict_with_fallback(self.udl_graph.parents)
        # --- set a flag used to determine if the GraphScope is being used as a
        #     context manager
        self.__active = False
        # --- keep track of changed nodes
        self.changes = set() 

    # -------------------------------------------------------------------------
    def __enter__(self):
        self.__active = True
        # --- make sure the underlying graph hasn't changed since the graph
        #     scope was created
        if id(self.udl_graph) != id(onyx_dg.active_graph):
            raise GraphError("Active instance of the "
                "DependencyGraph has changed since the GraphScope was created")
        # --- replace the active graph with the GraphScope
        onyx_dg.active_graph = self
        # --- recursively invalidate all ancestors of each changed node: to
        #     minimize invalidation first collect parents of all changed nodes
        parents = set()
        for changed in self.changes:
            parents.update(self.parents[changed])
        for parent in parents:
            self.invalidate_node(parent)
        # --- return a reference to itself (to be used for disposable scopes)
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        onyx_dg.active_graph = self.udl_graph
        self.__active = False
        # --- returns False so that all execptions raised will be propagated
        return False

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = super().__deepcopy__(memo)
        clone.__active = self.__active
        clone.changes = self.changes.copy()
        clone.udl_graph = self.udl_graph
        memo[id(self)] = clone
        return clone

    # -------------------------------------------------------------------------
    def change_value(self, obj, attr, value):
        """
        Description:
            Change the value of a graph node within a GraphScope so that the
            change and all its implications won't be visible outside the scope.
            NB: only Property and Settable (including stored attributes) nodes
                can be changed.
        Inputs:
            obj   - instance (or name) of an object derived from UfoBase
            attr  - name of the target attribute/method
            value - the new value for the attr
        Returns:
            None.
        """
        if not self.__active:
            # --- switch the active graph to the graph scope. also make sure
            #     that the underlying graph hasn't changed since the graph
            #     scope was created
            if id(self.udl_graph) != id(onyx_dg.active_graph):
                raise GraphError(
                    "Active instance of the DependencyGraph"
                    "has changed since the GraphScope was created")
            # --- replace the active graph with the GraphScope
            onyx_dg.active_graph = self

        try:
            name = obj if isinstance(obj, str) else obj.Name
            node_id = (name, attr, ())
            node = self.get_or_create(node_id)

            if not isinstance(node, (PropertyNode, SettableNode)):
                node_type = node.__class__.__name__
                raise GraphError(
                    "Unsupported node type: {0:s}".format(node_type))

            # --- keep track of changed nodes
            self.changes.add(node_id)

            # --- NB: this will also save a local copy of all affected nodes
            self.invalidate_node(node_id)

            # --- set the node value and set its state to valid
            node.value = value
            node.valid = True

        finally:
            if not self.__active:
                onyx_dg.active_graph = self.udl_graph
